"""Hub module for hosting tools."""

import importlib
import json
import os
from collections.abc import Callable
from types import ModuleType
from huggingface_hub import hf_hub_download
from huggingface_hub import snapshot_download
from huggingface_hub import upload_file
from pydantic import BaseModel


class AgentHub:
    """Hugging Face Repo class.

    This class provides an interface to interact with Hugging Face repositories.
    It allows loading files, configs, and tools from a specified repository,
    as well as handling local caching when possible.
    """

    tools_dir: str = "tools"
    agents_dir: str = "agents"
    pydantic_models_dir: str = "pydantic_models"
    repo_type: str = "space"

    def __init__(self, repo_id: str) -> None:
        """Initialize the Hugging Face Hub."""
        self.repo_id = repo_id

    def download_files(self) -> None:
        """Download all files from the Hugging Face Hub."""
        snapshot_download(
            repo_id=self.repo_id,
            repo_type=self.repo_type,
        )

    def get_file_path(self, filename: str, subdir: str) -> str:
        """Get the local path to a file in the repo."""
        match subdir:
            case self.tools_dir:
                extension = ".py"
                subfolder = self.tools_dir
            case self.pydantic_models_dir:
                extension = ".py"
                subfolder = self.pydantic_models_dir
            case self.agents_dir:
                extension = ".json"
                subfolder = self.agents_dir
            case _:  # pragma: no cover
                raise ValueError(f"Invalid subdir: {subdir}")

        filename = self._check_extension(filename, extension)
        file_path = hf_hub_download(
            repo_id=self.repo_id,
            filename=filename,
            subfolder=subfolder,
            local_files_only=True,
            repo_type=self.repo_type,
        )
        return file_path

    def _load_module(self, module_name: str, path: str) -> ModuleType:
        """Load a module from a path."""
        spec = importlib.util.spec_from_file_location(module_name, path)
        if spec is None or spec.loader is None:  # pragma: no cover
            raise ImportError(f"Could not load spec for module {path}")
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module

    def _check_extension(self, filename: str, extension: str) -> str:
        """Check if the filename has the correct extension."""
        if not filename.endswith(extension):  # pragma: no cover
            filename = f"{filename}{extension}"
        return filename

    def load_config(self, filename: str) -> dict:
        """Load a config from the Hugging Face Hub."""
        if not filename:  # pragma: no cover
            return {}
        file_path = self.get_file_path(filename, self.agents_dir)

        with open(file_path) as file:
            return json.load(file)

    def load_tool(self, filename: str) -> Callable | None:
        """Load a tool from the Hugging Face Hub."""
        if not filename:  # pragma: no cover
            return None
        name_without_extension = filename.split(".")[0]
        file_path = self.get_file_path(filename, self.tools_dir)
        module = self._load_module(name_without_extension, file_path)
        func = getattr(module, name_without_extension)
        assert callable(func)

        return func

    def load_structured_output(self, filename: str) -> type[BaseModel] | None:
        """Load a structured output from the Hugging Face Hub."""
        if not filename:  # pragma: no cover
            return None
        name_without_extension = filename.split(".")[0]
        file_path = self.get_file_path(filename, self.pydantic_models_dir)
        module = self._load_module(name_without_extension, file_path)
        model = getattr(module, name_without_extension)
        assert isinstance(model, type) and issubclass(model, BaseModel)

        return model

    def upload_content(
        self,
        filename: str,
        content: str,
        subdir: str,
    ) -> None:
        """Upload a file to the Hugging Face Hub."""
        match subdir:
            case self.tools_dir:
                extension = ".py"
            case self.pydantic_models_dir:
                extension = ".py"
            case self.agents_dir:
                extension = ".json"
            case _:
                raise ValueError(f"Invalid type: {subdir}")

        filename = self._check_extension(filename, extension)

        upload_file(
            path_or_fileobj=content.encode("utf-8"),
            path_in_repo=f"{subdir}/{filename}",
            repo_id=self.repo_id,
            repo_type=self.repo_type,
            commit_message=f"Update {filename}",
        )

    def list_files(self, subdir: str) -> list[str]:
        """List all files in the Hugging Face Hub."""
        repo_dir = snapshot_download(
            repo_id=self.repo_id, repo_type=self.repo_type, local_files_only=True
        )
        subdir_path = os.path.join(repo_dir, subdir)

        if not os.path.exists(subdir_path):
            return []

        if not os.path.isdir(subdir_path):
            return []

        return [
            f.split(".")[0]
            for f in os.listdir(subdir_path)
            if os.path.isfile(os.path.join(subdir_path, f))
        ]



from collections.abc import Callable
from unittest.mock import Mock
from unittest.mock import mock_open
from unittest.mock import patch
import pytest
from pydantic import BaseModel
from aic_core.agent_hub import AgentHub


# Test fixtures and helper classes
class DummyModel(BaseModel):
    name: str
    value: int


def dummy_function():
    pass


@pytest.fixture
def agent_hub():
    return AgentHub("test-repo")


# Tests
def test_init():
    repo = AgentHub("test-repo")
    assert repo.repo_id == "test-repo"


@patch("aic_core.agent_hub.snapshot_download")
def test_load_files(mock_snapshot):
    repo = AgentHub("test-repo")
    repo.download_files()
    mock_snapshot.assert_called_once_with(repo_id="test-repo", repo_type="space")


@patch("aic_core.agent_hub.hf_hub_download")
def test_load_config(mock_download):
    repo = AgentHub("test-repo")
    mock_download.return_value = "config.json"

    with patch("builtins.open", mock_open(read_data='{"key": "value"}')):
        result = repo.load_config("config")
        assert result == {"key": "value"}


@patch("aic_core.agent_hub.hf_hub_download")
@patch("aic_core.agent_hub.importlib.util")
def test_load_tool(mock_importlib, mock_download):
    repo = AgentHub("test-repo")
    mock_download.return_value = "/path/to/tool.py"

    # Setup mock module
    mock_module = Mock()
    mock_module.tool = dummy_function
    mock_spec = Mock()
    mock_spec.loader = Mock()

    mock_importlib.spec_from_file_location.return_value = mock_spec
    mock_importlib.module_from_spec.return_value = mock_module

    result = repo.load_tool("tool")
    assert isinstance(result, Callable)


@patch("aic_core.agent_hub.hf_hub_download")
@patch("aic_core.agent_hub.importlib.util")
def test_load_structured_output(mock_importlib, mock_download):
    repo = AgentHub("test-repo")
    mock_download.return_value = "/path/to/model.py"

    # Setup mock module
    mock_module = Mock()
    mock_module.model = DummyModel
    mock_spec = Mock()
    mock_spec.loader = Mock()

    mock_importlib.spec_from_file_location.return_value = mock_spec
    mock_importlib.module_from_spec.return_value = mock_module

    result = repo.load_structured_output("model")
    assert issubclass(result, BaseModel)


def test_upload_content():
    # Initialize the repo
    repo = AgentHub("test-repo")

    test_cases = [
        # (filename, content, subdir, expected_extension)
        ("test_tool", "def test_tool(): pass", "tools", ".py"),
        (
            "test_model",
            "from pydantic import BaseModel\nclass test_model(BaseModel): pass",
            "pydantic_models",
            ".py",
        ),
        ("test_config", '{"key": "value"}', "agents", ".json"),
    ]

    for filename, content, subdir, extension in test_cases:
        with patch("aic_core.agent_hub.upload_file") as mock_upload:
            # Call the method
            repo.upload_content(filename, content, subdir)

            # Verify upload_file was called with correct arguments
            mock_upload.assert_called_once_with(
                path_or_fileobj=content.encode("utf-8"),
                path_in_repo=f"{subdir}/{filename}{extension}",
                repo_id=repo.repo_id,
                repo_type=repo.repo_type,
                commit_message=f"Update {filename}{extension}",
            )


def test_upload_content_invalid_subdir():
    repo = AgentHub("test-repo")

    with pytest.raises(ValueError, match="Invalid type: invalid_dir"):
        repo.upload_content("test", "content", "invalid_dir")


def test_upload_content_with_extension():
    repo = AgentHub("test-repo")

    with (
        patch("aic_core.agent_hub.upload_file") as mock_upload,
    ):
        # Call with filename that already has extension
        repo.upload_content("test_tool.py", "content", "tools")

        # Verify correct handling of existing extension
        mock_upload.assert_called_once_with(
            path_or_fileobj=b"content",
            path_in_repo="tools/test_tool.py",
            repo_id=repo.repo_id,
            repo_type=repo.repo_type,
            commit_message="Update test_tool.py",
        )


def test_list_files_existing_directory():
    """Test listing files in an existing directory."""
    with (
        patch("aic_core.agent_hub.snapshot_download") as mock_snapshot,
        patch("os.path.exists") as mock_exists,
        patch("os.path.isdir") as mock_isdir,
        patch("os.listdir") as mock_listdir,
        patch("os.path.isfile") as mock_isfile,
    ):
        # Setup mocks
        mock_snapshot.return_value = "/fake/repo/path"
        mock_exists.return_value = True
        mock_isdir.return_value = True
        mock_listdir.return_value = ["file1.py", "file2.py", "file3.json"]
        mock_isfile.return_value = True

        hub = AgentHub("test-repo")
        files = hub.list_files("tools")

        # Verify results
        assert files == ["file1", "file2", "file3"]
        mock_snapshot.assert_called_once_with(
            repo_id=hub.repo_id, repo_type=hub.repo_type, local_files_only=True
        )


def test_list_files_nonexistent_directory():
    """Test listing files in a non-existent directory."""
    with (
        patch("aic_core.agent_hub.snapshot_download") as mock_snapshot,
        patch("os.path.exists") as mock_exists,
    ):
        mock_snapshot.return_value = "/fake/repo/path"
        mock_exists.return_value = False

        hub = AgentHub("test-repo")
        files = hub.list_files("nonexistent")

        assert files == []


def test_list_files_not_a_directory():
    """Test listing files when path exists but is not a directory."""
    with (
        patch("aic_core.agent_hub.snapshot_download") as mock_snapshot,
        patch("os.path.exists") as mock_exists,
        patch("os.path.isdir") as mock_isdir,
    ):
        mock_snapshot.return_value = "/fake/repo/path"
        mock_exists.return_value = True
        mock_isdir.return_value = False

        hub = AgentHub("test-repo")
        files = hub.list_files("not_a_dir")

        assert files == []



from typing import Union
from unittest.mock import Mock
from unittest.mock import patch
import pytest
from huggingface_hub.errors import LocalEntryNotFoundError
from pydantic_ai import Agent
from pydantic_ai import Tool
from aic_core.agent import AgentConfig
from aic_core.agent import AgentFactory
from aic_core.agent import MCPServerStdio


def test_agent_config_initialization():
    """Test basic initialization of AgentConfig."""
    config = AgentConfig(
        model="openai:gpt-4o",
        name="TestAgent",
        system_prompt="Test prompt",
        retries=2,
        repo_id="test-repo",
    )

    assert config.model == "openai:gpt-4o"
    assert config.name == "TestAgent"
    assert config.system_prompt == "Test prompt"
    assert config.retries == 2
    assert config.hf_tools == []
    assert config.mcp_servers == []


@patch("aic_core.agent.AgentHub")
def test_from_hub(mock_agent_hub):
    """Test loading config from Hugging Face Hub."""
    # Create a Mock instance for the repo
    mock_repo = Mock()
    mock_agent_hub.return_value = mock_repo

    # Set up the load_config mock on the repo instance
    mock_repo.load_config.return_value = {
        "model": "openai:gpt-4o",
        "name": "TestAgent",
        "system_prompt": "Test prompt",
    }

    config = AgentConfig.from_hub("test-repo", "agent")

    assert config.model == "openai:gpt-4o"
    assert config.name == "TestAgent"
    assert config.system_prompt == "Test prompt"
    mock_repo.load_config.assert_called_with("agent")


@patch("aic_core.agent.AgentHub")
def test_push_to_hub(mock_agent_hub):
    """Test pushing config to Hugging Face Hub."""
    mock_repo = Mock()
    mock_agent_hub.return_value = mock_repo

    config = AgentConfig(model="openai:gpt-4o", name="TestAgent", repo_id="test-repo")

    config.push_to_hub()

    mock_repo.upload_content.assert_called_once()


@pytest.fixture
def basic_config():
    return AgentConfig(
        model="openai:gpt-4o",
        name="TestAgent",
        result_type=["str"],
        known_tools=["tool1"],
        hf_tools=["tool2"],
        mcp_servers=["command1 arg1 arg2", "command2"],
        repo_id="test-repo",
    )


@pytest.fixture
def agent_factory(basic_config):
    return AgentFactory(basic_config)


def test_init(basic_config):
    factory = AgentFactory(basic_config)
    assert factory.config == basic_config


@patch("aic_core.agent.load_tool")
def test_hf_to_pai_tools_local(mock_load_tool):
    # Setup mock tool
    def forward():
        """Test tool docstring"""
        pass

    mock_tool = Mock()
    mock_tool.forward = forward
    mock_tool.name = "test_tool"
    mock_tool.description = "test description"
    mock_load_tool.return_value = mock_tool

    tool = AgentFactory.hf_to_pai_tools("test_tool")
    assert isinstance(tool, Tool)
    assert tool.name == "test_tool"

    # Verify local_files_only was tried first
    mock_load_tool.assert_called_with(
        "test_tool", trust_remote_code=True, local_files_only=True
    )


@patch("aic_core.agent.load_tool")
def test_hf_to_pai_tools_remote(mock_load_tool):
    # First call raises LocalEntryNotFoundError, second call succeeds
    def forward():
        """Test tool docstring"""
        pass

    mock_load_tool.side_effect = [
        LocalEntryNotFoundError("Not found locally"),
        Mock(
            forward=forward,
            name="test_tool",
            description="test description",
            forward__doc__=None,
        ),
    ]

    tool = AgentFactory.hf_to_pai_tools("test_tool")
    assert isinstance(tool, Tool)
    assert mock_load_tool.call_count == 2


def test_get_result_type_empty(agent_factory):
    agent_factory.config.result_type = []
    assert agent_factory.get_result_type() == str  # noqa: E721


def test_get_result_type_single(agent_factory):
    agent_factory.config.result_type = ["str"]
    assert agent_factory.get_result_type() == str  # noqa: E721


@patch("aic_core.agent.AgentHub")
def test_get_result_type_structured_output(mock_agent_hub, agent_factory):
    # Setup mock for AgentHub
    mock_repo = Mock()
    mock_custom_type = type("CustomType", (), {})  # Creates a dynamic type
    mock_repo.load_structured_output.return_value = mock_custom_type
    mock_agent_hub.return_value = mock_repo

    # Test with a mix of built-in and custom types
    agent_factory.config.result_type = ["str", "custom_output_type"]
    result = agent_factory.get_result_type()

    # Verify AgentHub was called for custom type
    mock_repo.load_structured_output.assert_called_once_with("custom_output_type")

    # Verify the resulting Union type contains both str and our custom type
    assert result == Union.__getitem__((str, mock_custom_type))


@patch("aic_core.agent.AgentHub")
def test_get_tools(mock_agent_hub, agent_factory):
    # Setup mocks
    def example_tool():
        """Example tool docstring"""
        pass

    mock_repo = Mock()
    mock_repo.load_tool.return_value = example_tool
    mock_agent_hub.return_value = mock_repo

    with patch.object(AgentFactory, "hf_to_pai_tools") as mock_hf_to_pai:
        mock_hf_to_pai.return_value = Tool(example_tool, name="hf_tool")
        tools = agent_factory.get_tools()

        assert len(tools) == 2  # One known_tool and one hf_tool
        assert all(isinstance(tool, Tool) for tool in tools)


def test_get_mcp_servers(agent_factory):
    servers = agent_factory.get_mcp_servers()
    assert len(servers) == 2
    assert servers[0] == MCPServerStdio("command1", ["arg1", "arg2"])
    assert servers[1] == MCPServerStdio("command2", [])


@patch("aic_core.agent.OpenAIProvider")
def test_create_agent(mock_provider, agent_factory):
    # Setup mocks
    mock_provider_instance = Mock()
    mock_provider.return_value = mock_provider_instance

    with (
        patch.object(AgentFactory, "get_tools") as mock_get_tools,
        patch.object(AgentFactory, "get_result_type") as mock_get_result_type,
        patch.object(AgentFactory, "get_mcp_servers") as mock_get_mcp_servers,
    ):
        mock_get_tools.return_value = []
        mock_get_result_type.return_value = str
        mock_get_mcp_servers.return_value = []

        agent = agent_factory.create_agent("test-api-key")

        assert isinstance(agent, Agent)
        mock_provider.assert_called_once_with(api_key="test-api-key")



"""Agent module."""

from typing import Any
from typing import Union
from huggingface_hub.errors import LocalEntryNotFoundError
from pydantic import BaseModel
from pydantic import Field
from pydantic_ai import Agent
from pydantic_ai import Tool
from pydantic_ai.agent import ModelSettings
from pydantic_ai.mcp import MCPServerStdio
from pydantic_ai.models.openai import OpenAIModel
from pydantic_ai.providers.openai import OpenAIProvider
from smolagents import load_tool
from aic_core.agent_hub import AgentHub


class AgentConfig(BaseModel):
    """Configuration class for Agent initialization."""

    model: str
    """Model name. Must be a valid pydantic_ai.models.KnownModelName."""
    result_type: list[str] = ["str"]
    """Result type. Can be name of Python primitives or a list of HF Hub file names."""
    system_prompt: str = "You are a helpful assistant."
    """System prompt for the agent."""
    # deps_type: type[AgentDepsT] = (NoneType,)
    name: str = "Agent"
    """Name of the agent."""
    model_settings: dict | None = None
    """Model settings for the agent."""
    retries: int = 1
    """Number of retries for the agent."""
    result_tool_name: str = "final_result"
    """Name of the result tool."""
    result_tool_description: str | None = None
    """Description of the result tool."""
    result_retries: int | None = None
    """Number of retries for the result tool."""
    known_tools: list[str] = []
    """List of known tools for the agent."""
    hf_tools: list[str] = []
    """List of Hugging Face tools for the agent."""
    mcp_servers: list[str] = []
    """List of MCP servers commands for the agent."""
    defer_model_check: bool = False
    """Whether to defer model check for the agent."""
    end_strategy: str = "early"
    """End strategy for the agent."""
    config_version: str = "0.0.1"
    """Version of the agent config."""
    repo_id: str = Field()
    """Hugging Face repo ID."""

    @classmethod
    def from_hub(cls, repo_id: str, agent_name: str) -> "AgentConfig":
        """Load an agent config from Hugging Face Hub."""
        repo = AgentHub(repo_id)
        config_obj = repo.load_config(agent_name)
        return cls(**config_obj, repo_id=repo_id)

    def push_to_hub(self) -> None:
        """Upload a JSON file to Hugging Face Hub."""
        repo = AgentHub(self.repo_id)

        # Verify before uploading
        agent_factory = AgentFactory(self)
        agent_factory.create_agent(api_key="key")

        # Upload the file
        repo.upload_content(
            filename=self.name,
            content=self.model_dump_json(indent=2),
            subdir=AgentHub.agents_dir,
        )


class AgentFactory:
    """Factory class to create an agent from a config."""

    def __init__(self, config: AgentConfig):
        """Initialise the agent factory."""
        self.config = config

    @classmethod
    def hf_to_pai_tools(cls, tool_name: str) -> Tool:
        """Convert a Hugging Face tool to a Pydantic AI tool."""
        try:
            tool = load_tool(tool_name, trust_remote_code=True, local_files_only=True)
        except LocalEntryNotFoundError:
            tool = load_tool(tool_name, trust_remote_code=True)
        return Tool(
            tool.forward,
            name=tool.name,
            # Do nothing if the tool function already has a docstring
            description=tool.description if not tool.forward.__doc__ else None,
        )

    def get_result_type(self) -> Any:
        """Creates a Union type from a list of types."""
        if not self.config.result_type:
            return str
        if len(self.config.result_type) == 1:
            return eval(self.config.result_type[0])

        type_objects = []
        for type_str in self.config.result_type:
            # Handle built-in types and local class types
            try:
                type_objects.append(eval(type_str))
            except NameError:  # Structured output
                hf_repo = AgentHub(self.config.repo_id)
                type_objects.append(hf_repo.load_structured_output(type_str))

        # Create a new type using Union
        return Union.__getitem__(tuple(type_objects))

    def get_tools(self) -> list[Tool]:
        """Get the tools from known tools and hf tools."""
        tools = []
        hf_repo = AgentHub(self.config.repo_id)
        for tool_name in self.config.known_tools:
            tool = hf_repo.load_tool(tool_name)
            tools.append(Tool(tool))  # type: ignore[arg-type]
        for tool_name in self.config.hf_tools:
            tools.append(self.hf_to_pai_tools(tool_name))
        return tools

    def get_mcp_servers(self) -> list[MCPServerStdio]:
        """Get the MCP servers from the config."""
        servers = []
        for server in self.config.mcp_servers:
            if not server.strip():  # pragma: no cover
                continue
            command, *args = server.split()
            servers.append(MCPServerStdio(command, args))
        return servers

    def create_agent(self, api_key: str) -> Agent:
        """Create an agent from a config."""
        result_type = self.get_result_type()
        model_name = self.config.model.split(":")[1]
        model = OpenAIModel(model_name, provider=OpenAIProvider(api_key=api_key))
        return Agent(
            model=model,
            result_type=result_type,
            system_prompt=self.config.system_prompt,
            name=self.config.name,
            model_settings=ModelSettings(**self.config.model_settings)
            if self.config.model_settings
            else None,
            retries=self.config.retries,
            result_tool_name=self.config.result_tool_name,
            result_tool_description=self.config.result_tool_description,
            result_retries=self.config.result_retries,
            tools=self.get_tools(),
            mcp_servers=self.get_mcp_servers(),  # type: ignore[arg-type]
            defer_model_check=self.config.defer_model_check,
            end_strategy=self.config.end_strategy,  # type: ignore[arg-type]
        )
